/**\
|**|	CPU_Meter.c
\**/

/**\
|**| ==============================================================================
|**| COMPILER DIRECTIVES
|**| ==============================================================================
\**/

#define kMaxTasks 64

#define SystemSevenOrLater		1
#define DEBUGSTRINGS 			1

/**\
|**| ==============================================================================
|**| INCLUDES
|**| ==============================================================================
\**/

#ifndef USE_PRECOMPILED_HEADER
#include "MyHeaders.i"
#endif

#include <Devices.h>
#include <Fonts.h>
#include <CodeFragments.h>
#include <Dialogs.h>
#include <DiskInit.h>
#include <LowMem.h>
#include <math.h>
#include <PLStringFuncs.h>
#include <Sound.h>
#include <Traps.h>
#include <Threads.h>

#include <stdio.h>
#include <string.h>

#include <Math64.h>

#include "Multiprocessing.h"
#include "MPInfoLib.h"

#include "Preferences.h"

#if __profile__
#include <Profiler.h>
#endif

/**\
|**| ==============================================================================
|**| TYPEDEFS, STRUCTS, DEFINES, ENUMS, ETC.
|**| ==============================================================================
\**/

#if DEBUGSTRINGS
#	define DEBUGSTR(x) DebugStr((x))
#	define LOGSTRING(b,s) if (b) do {DEBUGSTR(s);} while (false);
#else
#	define DEBUGSTR(x)
#	define LOGSTRING(b,s)
#endif

enum
{
	mAppleMenu = 128,
		iAboutBox = 1,
	mFileMenu = 129,
		iQuit = 1,
	mEditMenu = 130,
		iUndo = 1,
		iCut = 3,
		iCopy = 4,
		iPaste = 5,
		iClear = 6,
		iDuplicate = 8,
		iSelectAll = 9,
	mShow = 131,
		iProcID = 1,
		iTaskID = 2,
		iName = 3,
		iQueue = 4,
		iCPUTime = 5,
		iDelta = 6,
		iPercent = 7,
		iPreemptions = 8,
		iDeltaPreemptions = 9,
		iWeight = 10,
		iBar = 11,
		iTotals = 12,
	mInterval = 132,
		iPointFive = 1,
		iOne = 2,
		iTwo = 3,
		iFive = 4,
		iTen = 5,
		iTwenty = 6,
		iThirty = 7,
		iMinute = 8
};


UInt32 gIntervals[] = {30, 60, 120, 300, 600, 1200, 1800, 3600};

#define kTextSize 14

typedef struct MyTaskStruct
{
	MPProcessID		fProcID;
	MPTaskID		fTaskID;
	MPTaskInfo		fInfo;
	MPTaskInfo		fLast;
	float			fPercent;
	Boolean			fValid;
} MyTaskRec, * MyTaskPtr, ** MyTaskHdl;

#define RECT_WIDTH(r) ((r).right - (r).left)
#define RECT_HEIGHT(r) ((r).bottom - (r).top)

#ifndef MIN
#	define ABS(x)	((x) >= 0 ? (x) : -(x))
#	define MIN(a,b) ((a) < (b) ? (a) : (b))
#	define MAX(a,b) ((a) > (b) ? (a) : (b))
#	define PIN(a,b,c) MIN(MAX((a),(b)),(c))
#endif

/**\
|**| ==============================================================================
|**| EXPORTED FUNCTION PROTOTYPES
|**| ==============================================================================
\**/

/**\
|**| ==============================================================================
|**| PRIVATE FUNCTION PROTOTYPES
|**| ==============================================================================
\**/

static OSStatus Init_Mac(void);
static void Handle_Command(const long ms);
static void Handle_Event(const EventRecord* pEventPtr);
static void Handle_NullEvent(const EventRecord* pEventPtr);
static void Handle_MouseEvent(const EventRecord* pEventPtr);
static void Handle_KeyEvent(const char key,const SInt16 modifiers);
static void Handle_UpdateEvent(WindowPtr updateWindowP);
static void Handle_ActivateEvent(WindowPtr updateWindowP);
static void Handle_DiskEvent(const long message);
static void Handle_OSEvent(const long message);

static Boolean SetUp_MenuBar(void);
static void Adjust_MenuItems(void);

static void Handle_ContentClick(WindowPtr pWindowPtr,const EventRecord* pEventPtr);

static void Collect_Info(void);
static void Draw_Info(void);

static UInt64 AbsoluteToMilliseconds(const AbsoluteTime t);
static OSStatus save_prefs(void);

static void Zoom_Window(WindowPtr theWindow,
						const short zoomDir,
						const short hMax,
						const short vMax);
static void Grow_WindowGrid(WindowPtr theWindow);
static void Drag_WindowGrid(WindowPtr win,const Point pt);

/**\
|**| ==============================================================================
|**| PRIVATE GLOBALS
|**| ==============================================================================
\**/

static WindowPtr gWindowPtr = nil;
static Rect gWindowRect;
static Boolean gInBackGround = false;
static Boolean gQuitFlag = false;

static SInt32 gNextTicks = 0;

static MyTaskRec gMyTaskRecs[kMaxTasks];
static UInt32 gMyTaskRecCount = 0;

static const RGBColor 
	blackRGBColor 	=	{0x0000, 0x0000, 0x0000},
	blueRGBColor	=	{0x0000, 0x0000, 0xFFFF},
	ltBlueRGBColor 	= 	{0x7FFF, 0x7FFF, 0xFFFF},
	dkBlueRGBColor 	= 	{0x0000, 0x0000, 0x7FFF},
	redRGBColor 	=	{0xFFFF, 0x0000, 0x0000},
	ltGreenRGBColor = 	{0x7FFF, 0xFFFF, 0x7FFF},
	greenRGBColor 	= 	{0x0000, 0xFFFF, 0x0000},
	dkGreenRGBColor =	{0x0000, 0x7FFF, 0x0000},
	yellowRGBColor 	=	{0xFFFF, 0xFFFF, 0x0000},
	dkGrayRGBColor 	=	{0x4444, 0x4444, 0x5444},
	mdGrayRGBColor 	=	{0x8888, 0x8888, 0x9888},
	ltGrayRGBColor 	=	{0xaaaa, 0xaaaa, 0xbaaa},
	whiteRGBColor	=  	{0xFFFF, 0xFFFF, 0xFFFF};

static const RGBColor greenRGBColors[] = {
	{13107, 26214, 0},
	{39371, 54248, 15934},
	{46517, 57568, 16448},
	{52685, 60909, 28013},
	{56283, 65021, 30069},
	{60909, 65535, 47031}
};

static const RGBColor dkGreenRGBColors[] = {
	{6554, 131707, 0},
	{19686, 27124, 7967},
	{23259, 28784, 8224},
	{26343, 30455, 14007},
	{28142, 32511, 15035},
	{30455, 32768, 23516}
};

static const RGBColor grayRGBColors[] = {
	{21845, 21845, 21845},
	{34952, 34952, 34952},
	{43690, 43690, 43690},
	{48059, 48059, 48059},
	{56797, 56797, 56797},
	{61166, 61166, 61166}
};

static Preferences gPreferences;

static EventRecord gTheEvent;					// from the main event loop
static Boolean gHasColorQD = true;
static Point gGridSize = {8, 4};

/**\
|**| ==============================================================================
|**| PRIVATE FUNCTIONS
|**| ==============================================================================
\**/

void main(void)
{
#if __profile__
	if (!ProfilerInit(collectDetailed, bestTimeBase, 20, 5))
	{
#endif

		if (!Init_Mac() && SetUp_MenuBar())
		{
			do
			{
				InitCursor();
				WaitNextEvent(everyEvent, &gTheEvent, 60, nil);
				Handle_Event(&gTheEvent);
			} while (!gQuitFlag);
		}

#if __profile__
	}
	ProfilerDump("\pCPU_Meter.dump");
	ProfilerTerm();
#endif

}

/**\
|**|	Initialize toolboxes
\**/

static OSStatus Init_Mac(void)
{
	OSStatus error;
	SysEnvRec theWorld;

	//
	//	Test the computer to be sure we can do color.  
	//	If not we would crash, which would be bad.  
	//	If we cant run, just beep and exit.
	//

	error = SysEnvirons(1, &theWorld);
	if (theWorld.hasColorQD == false)
	{
		SysBeep(50);
		ExitToShell();							// If no color QD, we must leave.
	}

	MaxApplZone();
	InitGraf(&(qd.thePort));
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);

	//	InitContextualMenus();

	GetDateTime((unsigned long*) & qd.randSeed);

	gWindowPtr = nil;							// just to be sure

	gPreferences.fVersion = -1; // illegal value
	error = ReadPreferences(&gPreferences);
	if (error == noErr)
		gWindowRect = gPreferences.fWindowRect;

	if (gPreferences.fVersion != kPrefsVersion)
	{
		gWindowRect = qd.screenBits.bounds;
		gWindowRect.top += LMGetMBarHeight() + 1;
		InsetRect(&gWindowRect, 100, 100);

		gPreferences.fVersion = kPrefsVersion;
		gPreferences.fInterval = gIntervals[1];
		gPreferences.fShowProcID = false;
		gPreferences.fShowTaskID = false;
		gPreferences.fShowName = true;
		gPreferences.fShowQueue = true;
		gPreferences.fShowCPUTime = false;
		gPreferences.fShowDeltaCPU = true;
		gPreferences.fShowPercent = true;
		gPreferences.fShowPreemptions = false;
		gPreferences.fShowDeltaPreemptions = true;
		gPreferences.fShowWeight = true;
		gPreferences.fShowBar = true;
		gPreferences.fShowTotals = true;
		save_prefs();
	}

	//
	//	Make a new window for drawing in, and it must be a color window.  
	//	The window is full screen size, made smaller to make it more visible.
	//


	gWindowPtr = NewCWindow(nil, &gWindowRect, "\pCPU Meter", true, zoomDocProc, (WindowPtr) - 1, false, 0);

	OffsetRect(&gWindowRect, -gWindowRect.left, -gWindowRect.top);

	SetPort(gWindowPtr);						// set window to current graf port
	TextSize(kTextSize);						// smaller font for drawing.

	{
		short familyID;
		GetFNum("\pMonaco", &familyID);
		TextFont(familyID);
	}

	return noErr;
}

/**\
|**|	Setup menu bar
\**/

static Boolean SetUp_MenuBar(void)
{
	Boolean result = false;
	Handle mBar = GetNewMBar(128);				// handle to menu bar

	if (!ResError() && mBar)
	{
		SetMenuBar(mBar);
		AppendResMenu(GetMenuHandle(128), 'DRVR');
		DrawMenuBar();
		ReleaseResource(mBar);
		result = true;
	}
	return result;
}

/**\
|**|	Adjust menu items
\**/

static void Adjust_MenuItems(void)
{
	MenuHandle tMenuHdl = GetMenuHandle(mShow);
	if (tMenuHdl)
	{
		CheckItem(tMenuHdl, iProcID, gPreferences.fShowProcID);
		CheckItem(tMenuHdl, iTaskID, gPreferences.fShowTaskID);
		CheckItem(tMenuHdl, iName, gPreferences.fShowName);
		CheckItem(tMenuHdl, iQueue, gPreferences.fShowQueue);
		CheckItem(tMenuHdl, iCPUTime, gPreferences.fShowCPUTime);
		CheckItem(tMenuHdl, iDelta, gPreferences.fShowDeltaCPU);
		CheckItem(tMenuHdl, iPercent, gPreferences.fShowPercent);
		CheckItem(tMenuHdl, iPreemptions, gPreferences.fShowPreemptions);
		CheckItem(tMenuHdl, iDeltaPreemptions, gPreferences.fShowDeltaPreemptions);
		CheckItem(tMenuHdl, iWeight, gPreferences.fShowWeight);
		CheckItem(tMenuHdl, iBar, gPreferences.fShowBar);
		CheckItem(tMenuHdl, iTotals, gPreferences.fShowTotals);
	}

	tMenuHdl = GetMenuHandle(mInterval);
	if (tMenuHdl)
	{
		SInt32 index, count = CountMItems(tMenuHdl);
		SInt32 interval = gPreferences.fInterval;
		SInt32 value;
		Boolean first = true;

		for (index = 1; index <= count; index++)
		{
			value = gIntervals[index - 1];

			EnableItem(tMenuHdl, index);

			if (value < interval)
				CheckItem(tMenuHdl, index, false);
			else
			{
				CheckItem(tMenuHdl, index, first);
				if (first)
					gPreferences.fInterval = value;
				first = false;
			}
		}
	}
}


/**\
|**|	Do menu command
\**/

static void Handle_Command(const long ms)
{
	short menuID = ms >> 16,
	menuItem = ms & 0xFFFF;

	switch (menuID)
	{
		case mAppleMenu:
			switch (menuItem)
			{
				case iAboutBox:					// Bring up alert for About.
					SysBeep(15);
					break;
				default:						// All non-About items in this menu are DAs.
					{
						Str255 daName;
						GetMenuItemText(GetMenuHandle(menuID), menuItem, daName);
						OpenDeskAcc(daName);
					}
					break;
			}
			break;
		case mFileMenu:
			switch (menuItem)
			{
				case iQuit:
					gQuitFlag = true;
					break;
			}
			break;
		case mEditMenu:
			switch (menuItem)
			{
				case iUndo:
					break;
				case iCut:
					break;
				case iCopy:
					break;
				case iPaste:
					break;
				case iClear:
					break;
				case iDuplicate:
					break;
				case iSelectAll:
					break;
			}
			break;
		case mShow:
			switch (menuItem)
			{
				case iProcID:
					gPreferences.fShowProcID = !gPreferences.fShowProcID;
					break;
				case iTaskID:
					gPreferences.fShowTaskID = !gPreferences.fShowTaskID;
					break;
				case iName:
					gPreferences.fShowName = !gPreferences.fShowName;
					break;
				case iQueue:
					gPreferences.fShowQueue = !gPreferences.fShowQueue;
					break;
				case iCPUTime:
					gPreferences.fShowCPUTime = !gPreferences.fShowCPUTime;
					break;
				case iDelta:
					gPreferences.fShowDeltaCPU = !gPreferences.fShowDeltaCPU;
					break;
				case iPercent:
					gPreferences.fShowPercent = !gPreferences.fShowPercent;
					break;
				case iPreemptions:
					gPreferences.fShowPreemptions = !gPreferences.fShowPreemptions;
					break;
				case iDeltaPreemptions:
					gPreferences.fShowDeltaPreemptions = !gPreferences.fShowDeltaPreemptions;
					break;
				case iWeight:
					gPreferences.fShowWeight = !gPreferences.fShowWeight;
					break;
				case iBar:
					gPreferences.fShowBar = !gPreferences.fShowBar;
					break;
				case iTotals:
					gPreferences.fShowTotals = !gPreferences.fShowTotals;
					break;
			}
			gNextTicks = TickCount();			// make it update now
			save_prefs();
			break;
		case mInterval:
			switch (menuItem)
			{
				case iPointFive:
				case iOne:
				case iTwo:
				case iFive:
				case iTen:
				case iTwenty:
				case iThirty:
				case iMinute:
					gPreferences.fInterval = gIntervals[menuItem - 1];
					gNextTicks = TickCount();	// make it update now
					save_prefs();
				default:
					break;
			}
	}
}

/**\
|**|	Handle event
\**/

static void Handle_Event(const EventRecord* pEventPtr)
{
	switch (pEventPtr->what)
	{
		case nullEvent:							// 0
			LOGSTRING(0, "\p|Handle_Event-I-Debug, nullEvent.;g");
			Handle_NullEvent(pEventPtr);
			break;
		case mouseDown:							// 1
			LOGSTRING(0, "\p|Handle_Event-I-Debug, mouseDown.;g");
			Handle_MouseEvent(pEventPtr);
			break;
		case mouseUp:							// 2
			LOGSTRING(0, "\p|Handle_Event-I-Debug, mouseUp.;g");
			break;
		case keyDown:							// 3
			LOGSTRING(0, "\p|Handle_Event-I-Debug, keyDown.;g");
			goto dokey;
			break;
		case keyUp:								// 4
			LOGSTRING(0, "\p|Handle_Event-I-Debug, keyUp.;g");
			break;
		case autoKey:							// 5
			LOGSTRING(0, "\p|Handle_Event-I-Debug, autoKey.;g");
dokey:		Handle_KeyEvent((char)(pEventPtr->message & charCodeMask), pEventPtr->modifiers);
			break;
		case updateEvt:							// 6
			LOGSTRING(0, "\p|Handle_Event-I-Debug, updateEvt.;g");
			Handle_UpdateEvent((WindowPtr)pEventPtr->message);
			break;
		case diskEvt:							// 7
			LOGSTRING(0, "\p|Handle_Event-I-Debug, diskEvt.;g");
			Handle_DiskEvent(pEventPtr->message);
			break;
		case activateEvt:						// 8
			LOGSTRING(0, "\p|Handle_Event-I-Debug, activateEvt.;g");
			Handle_ActivateEvent((WindowPtr)pEventPtr->message);
			break;
		case osEvt:								// 15
			LOGSTRING(0, "\p|Handle_Event-I-Debug, osEvt.;g");
			Handle_OSEvent(pEventPtr->message);
			break;
		case kHighLevelEvent:					// 23
			LOGSTRING(0, "\p|Handle_Event-I-Debug, kHighLevelEvent.;g");
			AEProcessAppleEvent(pEventPtr);
			break;
		default:
			LOGSTRING(0, "\p|pHandleEvent-I-Debug, UnHandled Event.;g");
			break;
	}
}

/**\
|**|	Do null event
\**/

static void Handle_NullEvent(const EventRecord* pEventPtr)
{
	(pEventPtr);

	if (gWindowPtr)
	{
		SInt32 nowTicks = TickCount();
		GrafPtr savePort;

		GetPort(&savePort);
		SetPort(gWindowPtr);

		if (nowTicks > gNextTicks)
		{
			gNextTicks = nowTicks + gPreferences.fInterval;
			InvalRect(&gWindowRect);
		}

		SetPort(savePort);
	}
}

/**\
|**|	Do mousedown event
\**/

static void Handle_MouseEvent(const EventRecord* pEventPtr)
{
	WindowPtr window;
	short part;

	part = FindWindow(pEventPtr->where, &window);

	if (part != inContent)
		SetCursor(&qd.arrow);

	switch (part)
	{
		case inContent:
			Handle_ContentClick(window, pEventPtr);
			if (window != FrontWindow())
			{
				SelectWindow(window);
				//				HiliteWindows();
			}
			else
				break;

		case inDrag:
			Drag_WindowGrid(window, pEventPtr->where);
			save_prefs();
			break;

		case inGoAway:
			if (TrackGoAway(window, pEventPtr->where))
			{
				save_prefs();

				DisposeWindow(window);
				//				DisposeOneWindow(window,kClose);
				if (window == gWindowPtr)
					gWindowPtr = nil;
			}
			break;

		case inGrow:
			{
				Grow_WindowGrid(window);
				gWindowRect = gWindowPtr->portRect;
				save_prefs();
			}
			break;

		case inMenuBar:							// Process mouse menu command (if any).
			{
				long ms;

				Adjust_MenuItems();
				ms = MenuSelect(pEventPtr->where);
				if (ms)
					Handle_Command(ms);
				HiliteMenu(0);					// Unhighlight what MenuSelect hilited.
			}
			break;

		case inSysWindow:						// Let the system handle the mouseDown.
			SystemClick(pEventPtr, window);
			break;

		case inZoomIn:
		case inZoomOut:
			{
				Rect zoomLimits = qd.screenBits.bounds;

				Zoom_Window(window, part, RECT_WIDTH(zoomLimits), RECT_HEIGHT(zoomLimits));

				gWindowRect = gWindowPtr->portRect;
				save_prefs();
			}
			break;
		default:
			break;
	}
}

/**\
|**|	Do key events
\**/

static void Handle_KeyEvent(const char key,const SInt16 modifiers)
{
	if ((modifiers & cmdKey) != 0)
	{
		Adjust_MenuItems();
		//		HandleMenuCommand(MenuKey(key));
		Handle_Command(MenuKey(key));
		HiliteMenu(0);							// Unhighlight what MenuSelect hilited.
	}
}

/**\
|**|	Do update events
\**/

static void Handle_UpdateEvent(WindowPtr updateWindowP)
{
	GrafPtr savePort;

	GetPort(&savePort);
	SetPort(updateWindowP);
	BeginUpdate(updateWindowP);

	if (updateWindowP == gWindowPtr)
	{

		Collect_Info();

		//		EraseRect(&gWindowRect);

		Draw_Info();
	}

	UpdateControls(updateWindowP, updateWindowP->visRgn);

	DrawGrowIcon(updateWindowP);
	EndUpdate(updateWindowP);

	SetPort(savePort);
}

/**\
|**|	Do disk events
\**/

static void Handle_DiskEvent(const long message)
{
	Point dialogLocation = {100, 100};

	if ((message & 0xFFFF0000) != noErr)
	{
		DIBadMount(dialogLocation, message);
	}
}

/**\
|**|	Do Activate events
\**/

static void Handle_ActivateEvent(WindowPtr updateWindowP)
{
#pragma unused(updateWindowP)
}


/**\
|**|	Do OS events
\**/

static void Handle_OSEvent(const long message)
{
	if ((message >> 24) == suspendResumeMessage)
	{
		if ((message & resumeFlag) != 0)
			gInBackGround = false;
		else
			gInBackGround = true;
	}
	else if ((message >> 24) == mouseMovedMessage)
	{
		// mouse moved!
	}
}

/**\
|**|	Do mouse down in window content
\**/

static void Handle_ContentClick(WindowPtr pWindowPtr,const EventRecord* pEventPtr)
{
#pragma unused (pEventPtr)
	if (pWindowPtr != FrontWindow())
	{
		SelectWindow(pWindowPtr);
		//		HiliteWindows();
	}
	else
	{
		gNextTicks = TickCount();				// make it update now
		//		SysBeep(15);
	}
}


static void Collect_Info(void)
{
	OSStatus err = noErr;
	MPTaskID tMPTaskID = kInvalidID;
	MPProcessID tMPProcID = kInvalidID;

	while (true)
	{
		SInt32 index;
		MPTaskInfo fInfo;
		fInfo.version = kMPTaskInfoVersion;

		// For every kernel process...
		err = MPGetNextProcessID(&tMPProcID);
		if (err != noErr)
			break;

		while (true)
		{
			// Iterate through all the tasks of the process.
			err = MPGetNextTaskID(tMPProcID, &tMPTaskID);
			if (err != noErr)
				break;

			err = MPExtractTaskState(tMPTaskID, kMPTaskStateTaskInfo, (void*) & fInfo);
			if (err != noErr)
				break;

			for (index = 0; index < gMyTaskRecCount; index++)
			{
				if ((tMPProcID == gMyTaskRecs[index].fProcID) &&
					(tMPTaskID == gMyTaskRecs[index].fTaskID))
					break;
				if (kInvalidID == gMyTaskRecs[index].fTaskID)
					break;
			}

			if (index >= gMyTaskRecCount)			// Didn't find it
			{
				index = gMyTaskRecCount;
				if (gMyTaskRecCount < kMaxTasks)
				{
					gMyTaskRecs[index].fTaskID = kInvalidID;
					gMyTaskRecCount++;
				}
			}

			if (index < gMyTaskRecCount)			// if we found a record...
			{
				// ... store new task record fInfo
				if (gMyTaskRecs[index].fTaskID == kInvalidID)
				{
					gMyTaskRecs[index].fProcID = tMPProcID;
					gMyTaskRecs[index].fTaskID = tMPTaskID;
					gMyTaskRecs[index].fInfo = fInfo;
				}
				gMyTaskRecs[index].fLast = gMyTaskRecs[index].fInfo;
				gMyTaskRecs[index].fInfo = fInfo;
				gMyTaskRecs[index].fValid = true;
			}
		}
	}
}


static void Draw_Info(void)
{
	Str255 tStr;
	Rect tRect = gWindowRect;
	UInt64 schedDelta, cpuTime, last_cpuTime, first_time, sched_time;
	UInt64 cpuTimeTotal = 0,schedDeltaTotal = 0;
	UInt32 preemptions,preemptionsTotal = 0;
	UInt32 deltaPreemptions,deltaPreemptionsTotal = 0;
	UInt32 weight,weightTotal = 0;
	float percent,percentage,percentTotal = 0;
	SInt32 index;

	//	tRect.bottom = tRect.top + (kTextSize * 2);
	tRect.bottom = tRect.top + 20;

	sprintf((Ptr)tStr, "   ");
	if (gPreferences.fShowProcID)
		sprintf((Ptr)tStr, "%s Proc ID ", tStr);

	if (gPreferences.fShowTaskID)
		sprintf((Ptr)tStr, "%s Task ID ", tStr);

	if (gPreferences.fShowName)
		sprintf((Ptr)tStr, "%sName ", tStr);

	if (gPreferences.fShowQueue)
		sprintf((Ptr)tStr, "%sQueue ", tStr);

	if (gPreferences.fShowCPUTime)
		sprintf((Ptr)tStr, "%s    cpuTime   ", tStr);

	if (gPreferences.fShowDeltaCPU)
		sprintf((Ptr)tStr, "%s  CPU  ", tStr);

	if (gPreferences.fShowPercent)
		sprintf((Ptr)tStr, "%s  cpu %% ", tStr);

	if (gPreferences.fShowPreemptions)
		sprintf((Ptr)tStr, "%s  Preemptions ", tStr);

	if (gPreferences.fShowDeltaPreemptions)
		sprintf((Ptr)tStr, "%s Preemp ", tStr);

	if (gPreferences.fShowWeight)
		sprintf((Ptr)tStr, "%s  Weight", tStr);

	C2PStr((Ptr)tStr);
	RGBForeColor(&blackRGBColor);
	RGBBackColor(&grayRGBColors[5]);

	TETextBox(&tStr[1], tStr[0], &tRect, teFlushLeft);

	OffsetRect(&tRect, 0, 21);

	for (index = 0; index < gMyTaskRecCount; index++)
	{
		if (!gMyTaskRecs[index].fValid)
			gMyTaskRecs[index].fTaskID = kInvalidID;
		else if (gMyTaskRecs[index].fTaskID != kInvalidID)
		{
			RGBForeColor(&blackRGBColor);
			RGBBackColor(&grayRGBColors[5]);

			// creationTime time
			first_time = AbsoluteToMilliseconds(gMyTaskRecs[index].fInfo.creationTime);

			// current total CPU time
			cpuTime = AbsoluteToMilliseconds(gMyTaskRecs[index].fInfo.cpuTime);

			// last total CPU time
			last_cpuTime = AbsoluteToMilliseconds(gMyTaskRecs[index].fLast.cpuTime);

			// current scheduled CPU time
			sched_time = AbsoluteToMilliseconds(gMyTaskRecs[index].fInfo.schedTime);

			// Compute this interval's CPU percentage based on the last scheduled time.
			// If the task has not acquired any scheduling intervals, the CPU% is zero.

			schedDelta = AbsoluteToMilliseconds(gMyTaskRecs[index].fInfo.schedTime) - AbsoluteToMilliseconds(gMyTaskRecs[index].fLast.schedTime);

			if (schedDelta > 0)
				percent = ((cpuTime - last_cpuTime) * 100.0) / schedDelta;// instantaneous CPU %
			else
				percent = 0.0;

			if (sched_time > first_time)
				percentage = 100.0 * cpuTime / (sched_time - first_time);
			else
				percentage = 0.0;

			// We may have some error in our sampling, so pin it between 0 & 100 %.
			percent = fmax(fmin(percent, 100.0), 0.0);
			percentage = fmax(fmin(percentage, 100.0), 0.0);

			// sprintf((Ptr) tStr,	"  ");

			tStr[0] = (gMyTaskRecs[index].fInfo.runState == kMPTaskRunning) ? '*' : ' ';
			tStr[1] = '\0';
			sprintf((Ptr)tStr, "%s%1d ", tStr, gMyTaskRecs[index].fInfo.lastCPU);

			if (gPreferences.fShowProcID)
				sprintf((Ptr)tStr, "%s%08lx ", tStr, gMyTaskRecs[index].fProcID);

			if (gPreferences.fShowTaskID)
				sprintf((Ptr)tStr, "%s%08lx ", tStr, gMyTaskRecs[index].fTaskID);

			if (gPreferences.fShowName)
				sprintf((Ptr)tStr, "%s%.4s ", tStr, &gMyTaskRecs[index].fInfo.name);

			if (gPreferences.fShowQueue)
				sprintf((Ptr)tStr, "%s%.4s  ", tStr, &gMyTaskRecs[index].fInfo.queueName);

			if (gPreferences.fShowCPUTime)
				sprintf((Ptr)tStr, "%s%3lld:%02lld:%02lld.%03lld ", tStr, cpuTime / (1000 * 60 * 60),// hours
						(cpuTime / (1000 * 60)) % 60,// minutes
						(cpuTime / 1000) % 60,	// seconds
						cpuTime % 1000);		// thousandths

			if (gPreferences.fShowDeltaCPU)
				sprintf((Ptr)tStr, "%s%6lld ", tStr, schedDelta);

			if (gPreferences.fShowPercent)
				sprintf((Ptr)tStr, "%s %5.1f%% ", tStr, percent);

			preemptions = gMyTaskRecs[index].fInfo.preemptions;
			if (gPreferences.fShowPreemptions)
				sprintf((Ptr)tStr, "%s     %8ld ", tStr, preemptions);

			deltaPreemptions = gMyTaskRecs[index].fInfo.preemptions - gMyTaskRecs[index].fLast.preemptions;
			if (gPreferences.fShowDeltaPreemptions)
				sprintf((Ptr)tStr, "%s %6ld ", tStr, deltaPreemptions);

			weight = gMyTaskRecs[index].fInfo.weight;
			if (gPreferences.fShowWeight)
				sprintf((Ptr)tStr, "%s  %6ld ", tStr, weight);

			C2PStr((Ptr)tStr);
			TETextBox(&tStr[1], tStr[0], &tRect, teFlushLeft);

			while (gPreferences.fShowBar)
			{
				float oldPercent = gMyTaskRecs[index].fPercent;
				Rect tRect1 = tRect, tRect2, tRect3;
				UInt32 width;

				tRect1.left += StringWidth(tStr) + 4;
				tRect1.right -= 20;
				InsetRect(&tRect1, 0, 2);

				if (tRect1.left >= tRect1.right)
					break;

				tRect1.bottom = tRect1.top + 11;

				RGBForeColor(&blackRGBColor);
				RGBBackColor(&grayRGBColors[3]);

				FrameRect(&tRect1);
				InsetRect(&tRect1, 1, 1);

				width = tRect1.right - tRect1.left;

				tRect3 = tRect2 = tRect1;

				tRect2.left = tRect1.right = tRect1.left + (width * percent / 100.0f);
				tRect3.left = tRect2.right = tRect1.left + (width * fmax(percent, oldPercent) / 100.0f);

				{
					UInt32 index = 0;
					Rect tRect4 = tRect1;
					while (tRect4.top < tRect4.bottom)
					{
						RGBForeColor(&greenRGBColors[index++]);
						FrameRect(&tRect4);
						InsetRect(&tRect4, 1, 1);
					}
				}

				if (percent < oldPercent)
				{

					UInt32 index = 0;
					while (tRect2.top < tRect2.bottom)
					{
						RGBForeColor(&dkGreenRGBColors[index++]);
						FrameRect(&tRect2);
						InsetRect(&tRect2, 1, 1);
					}
				}

				{
					UInt32 index;
					for (index = 0; index < 6; index++)
					{
						RGBForeColor(&grayRGBColors[index]);
						FrameRect(&tRect3);
						InsetRect(&tRect3, 1, 1);
					}
					PaintRect(&tRect3);
				}

				gMyTaskRecs[index].fPercent = percent;

				tRect1.right = (tRect1.left += (width * percentage / 100)) + 2;

				RGBForeColor(&redRGBColor);
				PaintRect(&tRect1);

				break;
			}

			OffsetRect(&tRect, 0, 21);

			gMyTaskRecs[index].fValid = false;

			cpuTimeTotal += cpuTime;
			schedDeltaTotal += schedDelta;
			percentTotal += percent;
			preemptionsTotal += preemptions;
			deltaPreemptionsTotal += deltaPreemptions;
			weightTotal += weight;
		}
	}

	if (gPreferences.fShowTotals)
	{
		RGBForeColor(&blackRGBColor);
		RGBBackColor(&grayRGBColors[5]);

		sprintf((Ptr)tStr, "   ");	// 3

		if (gPreferences.fShowProcID)
			sprintf((Ptr)tStr, "%s         ", tStr);

		if (gPreferences.fShowTaskID)
			sprintf((Ptr)tStr, "%s         ", tStr);

		if (gPreferences.fShowName)
			sprintf((Ptr)tStr, "%s     ", tStr);

		if (gPreferences.fShowQueue)
			sprintf((Ptr)tStr, "%s      ", tStr);

		if (gPreferences.fShowCPUTime)
			sprintf((Ptr)tStr, "%s============= ", tStr);

		if (gPreferences.fShowDeltaCPU)
			sprintf((Ptr)tStr, "%s======= ", tStr);

		if (gPreferences.fShowPercent)
			sprintf((Ptr)tStr, "%s======= ", tStr);

		if (gPreferences.fShowPreemptions)
			sprintf((Ptr)tStr, "%s============= ", tStr);

		if (gPreferences.fShowDeltaPreemptions)
			sprintf((Ptr)tStr, "%s======== ", tStr);

		if (gPreferences.fShowWeight)
			sprintf((Ptr)tStr, "%s======== ", tStr);

		C2PStr((Ptr)tStr);
		TETextBox(&tStr[1], tStr[0], &tRect, teFlushLeft);

		OffsetRect(&tRect, 0, 21);

		sprintf((Ptr)tStr, "   ", tStr);

		if (gPreferences.fShowProcID)
			sprintf((Ptr)tStr, "%s         ", tStr);

		if (gPreferences.fShowTaskID)
			sprintf((Ptr)tStr, "%s         ", tStr);

		if (gPreferences.fShowName)
			sprintf((Ptr)tStr, "%s     ", tStr);

		if (gPreferences.fShowQueue)
			sprintf((Ptr)tStr, "%s      ", tStr);

		if (gPreferences.fShowCPUTime)
			sprintf((Ptr)tStr, "%s%3lld:%02lld:%02lld.%03lld ", tStr, cpuTimeTotal / (1000 * 60 * 60),// hours
					(cpuTimeTotal / (1000 * 60)) % 60,// minutes
					(cpuTimeTotal / 1000) % 60,	// seconds
					cpuTimeTotal % 1000);		// thousandths

		if (gPreferences.fShowDeltaCPU)
			sprintf((Ptr)tStr, "%s%6lld ", tStr, schedDeltaTotal);

		if (gPreferences.fShowPercent)
			sprintf((Ptr)tStr, "%s %5.1f%% ", tStr, percentTotal);

		if (gPreferences.fShowPreemptions)
			sprintf((Ptr)tStr, "%s     %8ld ", tStr, preemptionsTotal);

		if (gPreferences.fShowDeltaPreemptions)
			sprintf((Ptr)tStr, "%s %6ld ", tStr, deltaPreemptionsTotal);

		if (gPreferences.fShowWeight)
			sprintf((Ptr)tStr, "%s  %6ld ", tStr, weightTotal);

		C2PStr((Ptr)tStr);
		TETextBox(&tStr[1], tStr[0], &tRect, teFlushLeft);
		OffsetRect(&tRect, 0, 21);
	}

	RGBBackColor(&grayRGBColors[5]);
	tRect.bottom = gWindowRect.bottom;
	EraseRect(&tRect);
}

//*************************
//
// AbsoluteToMilliseconds
//
//*************************

static UInt64 AbsoluteToMilliseconds(const AbsoluteTime t)
{

	Nanoseconds nSec;
	UInt64 oneMillion = 1000000;

	nSec = AbsoluteToNanoseconds(t);

	return U64Div(UnsignedWideToUInt64(nSec), oneMillion);
}

/**\
|**|	save_prefs
\**/

static OSStatus save_prefs(void)
{
	gPreferences.fWindowRect = gWindowRect;
	if (gWindowPtr)
	{
		GrafPtr savePort;

		GetPort(&savePort);
		SetPort(gWindowPtr);

		LocalToGlobal((Point *) &gPreferences.fWindowRect.top);
		LocalToGlobal((Point *) &gPreferences.fWindowRect.bottom);

		SetPort(savePort);
	}
	return WritePreferences(&gPreferences);
}

/**\
|**|	Zoom_Window
\**/

static void Zoom_Window(WindowPtr theWindow,
						const short zoomDir,
						const short hMax,
						const short vMax)
{
	Rect * windRect, * zoomRect;
	Rect globalPortRect, theSect, dGDRect;
	GDHandle nthDevice, dominantGDevice;
	long sectArea,
	greatestArea;

	if (TrackBox(theWindow, gTheEvent.where, zoomDir))
	{
		SetPort(theWindow);
		EraseRect(&theWindow->portRect);		// recommended for cosmetic reasons

		if (zoomDir == inZoomOut)
		{
			/*
			 *	ZoomWindow() is a good basic tool, but it doesn't do everything necessary to
			 *	implement a good human interface when zooming. In fact it's not even close for
			 *	more high-end hardware configurations. We must help it along by calculating an
			 *	appropriate window size and location any time a window zooms out.
			 */

			windRect = &(**((WindowPeek)theWindow)->strucRgn).rgnBBox;
			dominantGDevice = nil;
			if (gHasColorQD)
			{

				/*
				 *	Color QuickDraw implies the possibility of multiple monitors. This is where
				 *	zooming becomes more interesting. One should zoom onto the monitor containing
				 *	the greatest portion of the window. This requires walking the gDevice list.
				 */

				nthDevice = GetDeviceList();
				greatestArea = 0;
				while (nthDevice != nil)
				{
					if (TestDeviceAttribute(nthDevice, screenDevice))
					{
						if (TestDeviceAttribute(nthDevice, screenActive))
						{
							SectRect(windRect, &(**nthDevice).gdRect, &theSect);
							sectArea = (long)RECT_WIDTH(theSect) * (long)RECT_HEIGHT(theSect);
							if (sectArea > greatestArea)
							{
								greatestArea = sectArea;// save the greatest intersection
								dominantGDevice = nthDevice;// and which device it belongs to
							}
						}
					}
					nthDevice = GetNextDevice(nthDevice);
				}
			}

			/*
			 *	At this point, we know the dimensions of the window we're zooming, and we know
			 *	what screen we're going to put it on. To be more specific, however, we need a
			 *	rectangle which defines the maximum dimensions of the resized window's contents.
			 *	This rectangle accounts for the thickness of the window frame, the menu bar, and
			 *	one or two pixels around the edges for cosmetic compatibility with ZoomWindow().
			 */

			if (dominantGDevice != nil)
			{
				dGDRect = (**dominantGDevice).gdRect;
				if (dominantGDevice == GetMainDevice())// account for menu bar on main device
					dGDRect.top += GetMBarHeight();
			}
			else
			{
				dGDRect = qd.screenBits.bounds;	// if no gDevice, use default monitor
				dGDRect.top += GetMBarHeight();
			}

			globalPortRect = theWindow->portRect;
			LocalToGlobal(&topLeft(globalPortRect));// calculate the window's portRect
			LocalToGlobal(&botRight(globalPortRect));// in global coordinates

			// account for the window frame and inset it a few pixels
			dGDRect.left += 2 + globalPortRect.left - windRect->left;
			dGDRect.top += 2 + globalPortRect.top - windRect->top;
			dGDRect.right -= 1 + windRect->right - globalPortRect.right;
			dGDRect.bottom -= 1 + windRect->bottom - globalPortRect.bottom;

			/*
			 *	Now we know exactly what our limits are, and since there are input parameters
			 *	specifying the dimensions we'd like to see, we can move and resize the zoom
			 *	state rectangle for the best possible results. We have three goals in this:
			 *	1. Display the window entirely visible on a single device.
			 *	2. Resize the window to best represent the dimensions of the document itself.
			 *	3. Move the window as short a distance as possible to achieve #1 and #2.
			 */

			zoomRect = &(**(WStateDataHandle)((WindowPeek)theWindow)->dataHandle).stdState;

			/*
			 *	Initially set the zoom rectangle to the size requested by the input parameters,
			 *	although not smaller than a minimum size. We do this without moving the origin.
			 */

			zoomRect->right = (zoomRect->left = globalPortRect.left) + hMax;
			//									MAX(hMax, MinWindowWidth(theWindow));
			zoomRect->bottom = (zoomRect->top = globalPortRect.top) + vMax;
			//									MAX(vMax, MinWindowHeight(theWindow));

			// Shift the entire rectangle if necessary to bring its origin inside dGDRect.
			OffsetRect(zoomRect, MAX(dGDRect.left - zoomRect->left, 0), MAX(dGDRect.top - zoomRect->top, 0));

			/*
			 *	Shift the rectangle up and/or to the left if necessary to accomodate the view,
			 *	and if it is possible to do so. The rectangle may not be moved such that its
			 *	origin would fall outside of dGDRect.
			 */

			OffsetRect(zoomRect, -PIN(zoomRect->right - dGDRect.right, 0, zoomRect->left - dGDRect.left), -PIN(zoomRect->bottom - dGDRect.bottom, 0, zoomRect->top - dGDRect.top));

			// Clip expansion to dGDRect, in case view is larger than dGDRect.
			zoomRect->right = MIN(zoomRect->right, dGDRect.right);
			zoomRect->bottom = MIN(zoomRect->bottom, dGDRect.bottom);
		}
		ZoomWindow(theWindow, zoomDir, false);	// all it needed was a brain transplant
	}
}

/**\
|**|	Grow_WindowGrid
\**/

#define kExtremeNeg -32768
#define kExtremePos (32767 - 1) // required to address an old region bug, see develop 20 Q&As

static void Pull_Rect(Rect* startRect);			// prototype
static void Grow_WindowGrid(WindowPtr theWindow)
{
	PenState oldPen;
	WindowPtr tempWP, WMPort;
	Rect draggingRect = theWindow->portRect;
	Rect wideOpen = {kExtremeNeg, kExtremeNeg, kExtremePos, kExtremePos};
	RgnHandle oldRgn = NewRgn();

	GetPort(&tempWP);
	SetPort(theWindow);

	// normalize my rectangle into the window manager port coordinates
	LocalToGlobal((Point *) & draggingRect.top);
	LocalToGlobal((Point *) & draggingRect.bottom);

	//  go to the WManager port
	GetWMgrPort(&WMPort);
	SetPort(WMPort);

	// save the Window manager pen state since we'll be changing it
	GetPenState(&oldPen);

	// localize back
	GlobalToLocal((Point *) & draggingRect.top);
	GlobalToLocal((Point *) & draggingRect.bottom);

	// save the current clip region, and set our wide-open clip
	GetClip(oldRgn);
	ClipRect(&wideOpen);

	// go to the routine below to do the actual dragging
	Pull_Rect(&draggingRect);

	// restore the original environment
	SetClip(oldRgn);
	SetPenState(&oldPen);
	DisposeRgn(oldRgn);

	// now size the window for the returned rect
	SetPort(theWindow);
	InvalRect(&theWindow->portRect);
	SizeWindow(theWindow, draggingRect.right - draggingRect.left, draggingRect.bottom - draggingRect.top, true);
	SetPort(tempWP);
}


static void Pull_Rect(Rect* startRect)
{
	Rect oldRect;
	Point endPoint;
	Boolean hreversed = false;
	Boolean vreversed = false;
	short tempH,
	tempV;
	short divByGridh,
	divByGridv;

	// set up
	oldRect = *startRect;
	PenMode(srcXor);							// So we can rubberband
	PenPat(&qd.gray);
	FrameRect(startRect);
	divByGridh = startRect->right;
	divByGridv = startRect->bottom;

	while (StillDown())
	{											// Keep doing this as long as the
		// user keeps the mouse down

		GetMouse(&endPoint);					// Current mouse position in local

		// see if it's on a  grid point
		tempH = ABS(endPoint.h - divByGridh) / gGridSize.h;
		tempV = ABS(endPoint.v - divByGridv) / gGridSize.v;

		// normalize to our grid values. We'll always go outwards as better
		if ((tempH * gGridSize.h) != ABS(endPoint.h - divByGridh))// shove out based on the remainer
			endPoint.h = (((endPoint.h) / gGridSize.h) * gGridSize.h) + gGridSize.h;
		if ((tempV * gGridSize.v) != ABS(endPoint.v - divByGridv))
			endPoint.v = (((endPoint.v) / gGridSize.v) * gGridSize.v) + gGridSize.v;

		// If things reversed, we have to make sure that we don't try
		// and grid the origin point of the drag, cuz that would be weird
		if (hreversed)
		{
			// see if the rectangle flipped first
			if (endPoint.h > startRect->right)	// they flipped back
				hreversed = false;				// and ignore this move
			else								// still reversed
				startRect->left = endPoint.h;
		}
		else
		{
			if (endPoint.h < startRect->left)
				hreversed = true;
			else
				startRect->right = endPoint.h;
		}
		if (vreversed)
		{
			// see if it flipped first
			if (endPoint.v > startRect->bottom)	// they flipped back
				vreversed = false;				// and ignore this move
			else								// still reversed
				startRect->top = endPoint.v;
		}
		else
		{
			if (endPoint.v < startRect->top)
				vreversed = true;
			else
				startRect->bottom = endPoint.v;
		}

		if (!EqualRect(startRect, &oldRect))
		{
			// redraw the rect only if the mouse moved
			FrameRect(&oldRect);
			FrameRect(startRect);				// draw the new rect
			oldRect = *startRect;
		}
	}
	FrameRect(startRect);

	PenMode(srcCopy);
	PenPat(&qd.black);
}


/*-------------------------------------------------------------------------------------

  Drag_WindowGrid- a big nasty function to Drag a window along grid lines.  
  Note the elegant error handling.  You should change it to make your users happy.

  You can change the size of the 'grid rects' by changing the value of kIncrement.

*/
static void Drag_WindowGrid(WindowPtr win,const Point pt)
{
	RgnHandle dragRgn, lastDragRgn, insetGray;
	GrafPtr oldPort, tmpPort;
	Point currPt, firstMulPoint, lastMulPoint, currMulPoint;
	Boolean frameHidden;
	enum {kIncrement = 16, kBorderInset = 4};

	// Set up our regions - init our variables.
	dragRgn = NewRgn();
	lastDragRgn = NewRgn();
	insetGray = NewRgn();
	if ((dragRgn == NULL) || (lastDragRgn == NULL) || (insetGray == NULL))
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}
	CopyRgn(GetGrayRgn(), insetGray);
	if (MemError() != noErr)
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}

	InsetRgn(insetGray, kBorderInset, kBorderInset);
	frameHidden = false;

	// Set up a port to draw into, and save off the old
	tmpPort = (GrafPtr)NewPtr(sizeof(GrafPort));
	if (tmpPort == NULL)
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}
	GetPort(&oldPort);
	OpenPort(tmpPort);
	CopyRgn(GetGrayRgn(), tmpPort->visRgn);
	if (MemError() != noErr)
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}
	tmpPort->portRect = (*GetGrayRgn())->rgnBBox;
	SetPort(tmpPort);

	PenMode(patXor);
	PenPat(&qd.gray);

	// Set the incoming point to be on a multiple of kIncrement,
	// to make later calculations easier.
	currMulPoint.h = pt.h + (kIncrement / 2);
	currMulPoint.h /= kIncrement;
	currMulPoint.h *= kIncrement;
	currMulPoint.v = pt.v + (kIncrement / 2);
	currMulPoint.v /= kIncrement;
	currMulPoint.v *= kIncrement;

	firstMulPoint = lastMulPoint = currMulPoint;

	CopyRgn(((WindowPeek)win)->strucRgn, dragRgn);
	if (MemError() != noErr)
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}

	CopyRgn(((WindowPeek)win)->strucRgn, lastDragRgn);
	if (MemError() != noErr)
	{
		DebugStr("\pnot enough memory- bye!");
		return;
	}

	// Draw the first framed region, which will follow the mouse
	// on the screen
	FrameRgn(lastDragRgn);

	while (WaitMouseUp() == true)
	{
		// Now track the mouse, and when it moves enough make the framed
		// region move as well.
		GetMouse(&currPt);

		// Set the new point to be on a multiple of kIncrement
		currMulPoint.h = currPt.h + (kIncrement / 2);
		currMulPoint.h /= kIncrement;
		currMulPoint.h *= kIncrement;
		currMulPoint.v = currPt.v + (kIncrement / 2);
		currMulPoint.v /= kIncrement;
		currMulPoint.v *= kIncrement;

		// Should we be showing the frame region ??
		if (PtInRgn(currPt, insetGray) == false)
		{
			// It's somewhere near the edges, so hide the frame
			if (frameHidden == false)			// if it's not hidden already, hide it now
				FrameRgn(lastDragRgn);
			frameHidden = true;
		}
		else
		{										// else, the frame should be shown if it's hidden
			if (frameHidden == true)
			{
				FrameRgn(lastDragRgn);
				frameHidden = false;
			}
		}

		// Has the mouse moved ?
		if (!EqualPt(currMulPoint, lastMulPoint))
		{
			// The mouse coordinates have changed enough to adjust the window,
			// so move the frame accordingly
			OffsetRgn(dragRgn, currMulPoint.h - lastMulPoint.h, currMulPoint.v - lastMulPoint.v);

			if (frameHidden == false)
			{
				// Only show the frame if we're allowed to.
				FrameRgn(dragRgn);
				FrameRgn(lastDragRgn);
			}
			lastMulPoint = currMulPoint;
			CopyRgn(dragRgn, lastDragRgn);
			if (MemError() != noErr)
			{
				DebugStr("\pnot enough memory- bye!");
				return;
			}
		}
	}

	// If frameHidden is true, there's no need to erase the final frame.
	if (frameHidden == false)
		FrameRgn(lastDragRgn);

	if (!EqualPt(lastMulPoint, firstMulPoint) && (frameHidden == false))
	{
		// The mouse has moved from its original position and is
		// somewhere on the screen, so move the window accordingly.
		Point globalPt, diffPt, contPt = {
										  0, 0};

		// Calculate the difference between the strucRgn's 0, 0 and
		// the window's content region 0, 0.  Remember that MoveWindow
		// moves the window's *content* to the coordinate specified, and
		// we want to move the window's structure to fit in the lastDragRgn
		SetPort(win);
		// LocalToGlobal works much better when the port is set up...
		LocalToGlobal(&contPt);
		SetPort(tmpPort);
		diffPt.h = contPt.h - (*((WindowPeek)win)->strucRgn)->rgnBBox.left;
		diffPt.v = contPt.v - (*((WindowPeek)win)->strucRgn)->rgnBBox.top;

		globalPt.h = (*lastDragRgn)->rgnBBox.left;
		globalPt.v = (*lastDragRgn)->rgnBBox.top;
		LocalToGlobal(&globalPt);
		globalPt.h += diffPt.h;
		globalPt.v += diffPt.v;
		MoveWindow(win, globalPt.h, globalPt.v, true);
	}

	// Close the port and tear everything down
	ClosePort(tmpPort);
	SetPort(oldPort);
	DisposeRgn(dragRgn);
	DisposeRgn(insetGray);
	DisposeRgn(lastDragRgn);
}
